use crate::context::Context;

pub struct Runner{
	lua:mlua::Lua,
}
#[derive(Debug)]
pub enum Error{
	Lua{
		source:String,
		error:mlua::Error
	},
	RustLua(mlua::Error),
	NoServices,
}
impl std::fmt::Display for Error{
	fn fmt(&self,f:&mut std::fmt::Formatter<'_>)->std::fmt::Result{
		match self{
			Self::Lua{source,error}=>write!(f,"lua error: source:\n{source}\n{error}"),
			Self::RustLua(error)=>write!(f,"rust-side lua error: {error}"),
			other=>write!(f,"{other:?}"),
		}
	}
}
impl std::error::Error for Error{}

fn init(lua:&mlua::Lua)->mlua::Result<()>{
	lua.sandbox(true)?;

	//global environment
	let globals=lua.globals();

	super::r#enum::set_globals(lua,&globals)?;
	super::color3::set_globals(lua,&globals)?;
	super::vector3::set_globals(lua,&globals)?;
	super::cframe::set_globals(lua,&globals)?;
	super::instance::set_globals(lua,&globals)?;

	Ok(())
}

impl Runner{
	pub fn new()->Result<Self,Error>{
		let runner=Self{
			lua:mlua::Lua::new(),
		};
		init(&runner.lua).map_err(Error::RustLua)?;
		Ok(runner)
	}
	pub fn runnable_context<'a>(self,context:&'a mut Context)->Result<Runnable<'a>,Error>{
		let services=context.find_services().ok_or(Error::NoServices)?;
		self.runnable_context_with_services(context,&services)
	}
	pub fn runnable_context_with_services<'a>(self,context:&'a mut Context,services:&crate::context::Services)->Result<Runnable<'a>,Error>{
		{
			let globals=self.lua.globals();
			globals.set("game",super::instance::DataModel::new(services.game)).map_err(Error::RustLua)?;
			globals.set("workspace",super::instance::Workspace::new(services.workspace)).map_err(Error::RustLua)?;
		}
		//this makes set_app_data shut up about the lifetime
		self.lua.set_app_data::<&'static mut rbx_dom_weak::WeakDom>(unsafe{core::mem::transmute(&mut context.dom)});
		Ok(Runnable{
			lua:self.lua,
			_lifetime:&std::marker::PhantomData
		})
	}
}

//Runnable is the same thing but has context set, which it holds the lifetime for.
pub struct Runnable<'a>{
	lua:mlua::Lua,
	_lifetime:&'a std::marker::PhantomData<()>
}
impl Runnable<'_>{
	pub fn drop_context(self)->Runner{
		self.lua.remove_app_data::<&'static mut rbx_dom_weak::WeakDom>();
		Runner{
			lua:self.lua,
		}
	}
	pub fn run_script(&self,script:super::instance::Script)->Result<(),Error>{
		let (name,source)=script.get_name_source(&self.lua).map_err(Error::RustLua)?;
		self.lua.globals().set("script",script).map_err(Error::RustLua)?;
		self.lua.load(source.as_str())
		.set_name(name)
		.exec().map_err(|error|Error::Lua{source,error})
	}
}
